#!/usr/bin/ruby -w
=begin
/***************************************************************************
*   Copyright (C) 2008, Paul Lutus                                        *
*   Changes - 2011 - by Al hart                                           *
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
*   This program is distributed in the hope that it will be useful,       *
*   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
*   GNU General Public License for more details.                          *
*                                                                         *
*   You should have received a copy of the GNU General Public License     *
*   along with this program; if not, write to the                         *
*   Free Software Foundation, Inc.,                                       *
*   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
***************************************************************************/

/***************************************************************************
*  Functions:                                                              *
*	$rb.beautify_directory(directory)                                            *
*	$rb.beautify_file(file)                                                *
*                                                                          *
* Options:  (See def initialize   )                                         *
*		@tab_str - space or tab                                            *
*		#@tab_size - number of spaces or tabe per tab                      *
*		@add_and_remove_blank_lines - and and remove excess blank lines    *
*                                                                          *
*                                                                          *
***************************************************************************/

=end

class RBeautify

	# these are regular expressions which cause an indent
	class CL_indent < RBeautify
		attr_accessor :term # term to add to end (e.g. end#def)
		attr_accessor :expr	# regular expression
		attr_accessor :new_line	# add a blank line before indent if there is not alreay a comment

		def initialize(term, new_line, expr)
			@term = term
			@new_line = new_line
			@expr = expr
			#printf("%s %s\n", term, expr)
		end#def#def

	end#class#classf

	# these are regular expressions which cause an outdent
	class CL_outdent < RBeautify
		attr_accessor :suppress	# supppress end#xxx modification (on else and elsif)
		attr_accessor :expr		# regular expression
		attr_accessor :new_line	# add blank line after outdent

		def initialize(suppress, new_line, expr)
			@suppress = suppress # if false, then add #xxx after outdent line (e.g.. end#if)
			@new_line = new_line # if true, add a new line after outdent line
			@expr = expr
			#printf("%s %s\n", term, expr)
		end#def#def

	end#class#class

	# user-customizable values

	def initialize
		#prograom options:
		# choose your tab character - spaces or tabs
		# comment out one pair
		#@tab_str = " "
		#@tab_size = 4
		@tab_str = "\t"
		@tab_size = 1

		#do you want to add blank lines before indents (e.g. class, def, if)
		# and after outdents
		# and limit consecutive blank lines to 2
		@add_and_remove_blank_lines = true

		@indent_expressions = [
			#			   term				new_line	expr
			CL_indent.new("module",			true,	/^module\b/),
			CL_indent.new("class",			true,	/^class\b/),
			CL_indent.new("if",				true,	/^if\b/),
			CL_indent.new("until",			true,	/(=\s*|^)until\b/),
			CL_indent.new("for",			true,	/(=\s*|^)for\b/),
			CL_indent.new("unless",			true,	/^unless\b/),
			CL_indent.new("while",			true,	/(=\s*|^)while\b/),
			CL_indent.new("begin",			true,	/(=\s*|^)begin\b/),
			CL_indent.new("case",			true,	/(^| )case\b/),
			CL_indent.new("then",			false,	/\bthen\b/),
			CL_indent.new("begin/rescue",	false,	/^rescue\b/),
			CL_indent.new("def",			true,	/^def\b/),
			CL_indent.new("do",				true,	/\bdo\b/),
			CL_indent.new("if",				false,	/^else\b/),
			CL_indent.new("if",				false,	/^elsif\b/),
			CL_indent.new("ensure",			false,	/^ensure\b/),
			CL_indent.new("when",			false,	/\bwhen\b/),
			CL_indent.new("brace",			false,	/\{[^\}]*$/),
			CL_indent.new("brace",			false,	/\[[^\]]*$/)
		]
		# outdent regexp tests
		@outdent_expressions = [
		#				   suppress			new_line	expr
			CL_outdent.new(true,			false,	/^rescue\b/ ),
			CL_outdent.new(true,			false,	/^ensure\b/ ),
			CL_outdent.new(true,			false,	/^elsif\b/ ),
			CL_outdent.new(false,			true,	/^end\b/ ),
			CL_outdent.new(true,			false,	/^else\b/ ),
			CL_outdent.new(true,			false,	/\bwhen\b/ ),
			CL_outdent.new(true,			false,	/^[^\{]*\}/ ),
			CL_outdent.new(true,			false,	/^[^\[]*\]/ )
		]
	end#def#def

	def rb_make_tab(tab_level)
		return (tab_level < 0)?"":@tab_str * @tab_size * tab_level
	end#def#def

	def rb_add_line(line,tab_level)
		line = rb_make_tab(tab_level) + line if line.length > 0
		return line
	end#def#def

	def beautify_string(source, path = "")
		comment_block = false
		in_here_doc = false
		here_doc_term = ""
		program_end = false
		multiLine_array = []
		multiLine_str = ""
		tab_level = 0
		@output = []
		tab_terms = []
		tab_lines = []
		lknt = 0
		@last_line_blank = false
		last_line_was_comment = false

		source.each_line do |line|
			line.strip!
			lknt = lknt + 1

			if(!program_end)
				# detect program end mark

				if(line =~ /^__END__$/)
					program_end = true
					printf("START program_end BLOCK %s\n", line)
				else
					# combine continuing lines

					if(!(line =~ /^\s*#/) && line =~ /[^\\]\\\s*$/)
						multiLine_array.push line
						multiLine_str += line.sub(/^(.*)\\\s*$/,"\\1")
						next
					end#if#if

					# add final line

					if(multiLine_str.length > 0)
						multiLine_array.push line
						multiLine_str += line.sub(/^(.*)\\\s*$/,"\\1")
					end#if#if

					tline = ((multiLine_str.length > 0) ? multiLine_str:line).strip

					if(tline =~ /^=begin/)
						comment_block = true
						printf("START COMMENT BLOCK %s\n", line)
					end#if#if

					if(in_here_doc)
						in_here_doc = false if tline =~ %r{\s*#{here_doc_term}\s*}
					else # not in here_doc

						if tline =~ %r{=\s*<<}
							here_doc_term = tline.sub(%r{.*=\s*<<-?\s*([_|\w]+).*},"\\1")
							in_here_doc = here_doc_term.size > 0

							if (here_doc_term.index("END_FILE"))
								in_here_doc = false # skip our .rbs stuff
							end#if#if

							if (in_here_doc)
								printf("IN_HERE_DOC: %s\n", line)
								printf("here_doc_term: %s\n", here_doc_term)
							end#if#if

						end#if#if

					end#if#if

				end#if#if

			end#if#if

			if(comment_block || program_end || in_here_doc)
				# add the line unchanged
				add_blank_line_after_this_outdent = false
				@output << line
			else
				comment_line = (tline =~ /^#/)

				if(!comment_line)
					# throw out sequences that will
					# only sow confusion

					while tline.gsub!(/\{[^\{]*?\}/,"")
					end#while#while

					while tline.gsub!(/\[[^\[]*?\]/,"")
					end#while#while

					while tline.gsub!(/'.*?'/,"")
					end#while#while

					while tline.gsub!(/".*?"/,"")
					end#while#while

					while tline.gsub!(/\`.*?\`/,"")
					end#while#while

					while tline.gsub!(/\([^\(]*?\)/,"")
					end#while#while

					while tline.gsub!(/\/.*?\//,"")
					end#while#while

					while tline.gsub!(/%r(.).*?\1/,"")
					end#while#while

					# delete end-of-line comments
					tline.sub!(/#[^\"]+$/,"")
					# convert quotes
					tline.gsub!(/\\\"/,"'")

					#printf("%s: %s %s\n", lknt, tab_level, line)
					if (tab_level > 0)

						@outdent_expressions.each do |oe|
							$oe = oe
							re = oe.expr

							if(tline =~ re)

								if (!oe.suppress && tab_terms[tab_level] != "")
									line.strip!

									if (!line.index('#'))
										$tab_level = tab_level
										$tab_terms = tab_terms
										$tab_lines = tab_lines
										line = line.strip + '#' + tab_terms[tab_level].to_s
									end#if#if

									# don't add blank lines after end#if
									if (oe.new_line && tab_terms[tab_level] != "if")
										add_blank_line_after_this_outdent = true
									end#if#if
								end#if#if
								tab_level -= 1
								break
							end#if#if
						end#do#do

					end#if#if

				end#if#if

				if(!comment_line)

					@indent_expressions.each do |ie|
						$ie = ie
						re = ie.expr

						if(tline =~ re && !(tline =~ /\s+end\s*$/))
							# add blank line before an indent
							# don't add a blank line after a comment
							if (@add_and_remove_blank_lines && ie.new_line && !last_line_was_comment)
								add_to_output("") # add blank line before new indent
							end#if#if
							break
						end#if#if
					end#do#do

				end#if#if

				if (comment_line)
					last_line_was_comment = true
				else
					last_line_was_comment = false
				end#if#if

				if (multiLine_array.length > 0)
					add_blank_line_after_this_outdent = false

					multiLine_array.each do |ml|
						@output << rb_add_line(ml,tab_level)
					end#do#do

					multiLine_array.clear
					multiLine_str = ""
				else
					add_to_output(rb_add_line(line,tab_level))

					if (add_blank_line_after_this_outdent)

						if (@add_and_remove_blank_lines)
							add_to_output("")
						end#if#if
						add_blank_line_after_this_outdent = false
					end#if#if
				end#if#if

				if(!comment_line)

					@indent_expressions.each do |ie|
						$ie = ie
						re = ie.expr

						if(tline =~ re && !(tline =~ /\s+end\s*$/))
							tab_level += 1
							tab_terms[tab_level] = ie.term
							tab_lines[tab_level] = lknt;
							#@output << sprintf("#%s. TAB: %s: %s: %s", lknt, tab_level, ie.expr, tab_terms[tab_level])
							#printf("%s. TAB: %s: %s: %s\n", lknt, tab_level, ie.expr, tab_terms[tab_level])
							#printf("TAB LINE: %s\n", line)
							break
						end#if#if
					end#do#do

				end#if#if
			end#if#if

			if(tline =~ /^=end/)
				comment_block = false
			end#if#if

		end#do#do

		printf("LKNT: %s\n", lknt)
		error = (tab_level != 0)
		UI.messagebox("Error: indent/outdent mismatch: #{tab_level}.") if error
		return @output.join("\n") + "\n",error
	end # beautify_string#def

	def add_to_output(line)

		if (@add_and_remove_blank_lines && line.strip == "")

			if (@last_line_blank)
				return # skip blank liness
			end#if#if
			@last_line_blank = true
		else
			@last_line_blank = false
		end#if#if
		@output << line
	end#if#def

	def beautify_file(path)
		printf("BEAUTIFY FILE: %s\n", path)
		error = false
		# named file source
		source = File.read(path)
		$source = source
		dest,error = beautify_string(source,path)
		$dest = dest

		if(source != dest)
			# make a backup copy
			File.open(path + ".orig","w") { |f| f.write(source) }
			# overwrite the original
			File.open(path,"w") { |f| f.write(dest) }
		else
			printf("NO CHANGE: %s\n", path)
		end#if#if

		return error
	end # beautify_file#def

	def beautify_directory(directory)

		Dir.entries( directory ).each do |file|
			next if file == "."
			next if file == ".."
			#printf("FILE: %s %s\n", directory, file)

			# delete any files
			if (!File.directory?( file ))
				spath =  File.join(directory,file)
				#printf("SPATH: %\ns", spath)

				self

				if (spath.slice(spath.length-3 , spath.length ).downcase == ".rb")
					#printf("PROCESS: %s\n", spath)
					$rb.beautify_file(spath)
				end#if#if
			end#if#if
		end#do#do

	end#def#def

end # module RBeautify#class

# start an instance of the class
$rb = RBeautify.new

# beautify a directory
#$rb.beautify_directory("c:/aa")

$rb.beautify_file("c:/aa/bomb.rb")

